#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom tensorflow layers.
"""
import tensorflow as tf
tf.keras.backend.set_floatx('float64')
tf.config.threading.set_inter_op_parallelism_threads(2)
tf.config.threading.set_intra_op_parallelism_threads(8)
tf.config.set_soft_device_placement(True)
from tensorflow.keras.layers import Layer


class InitializableDense(Layer):
	''' 
	'''
	def __init__(self, units, initial_w=None, initial_b=None, bias=False):
		'''
		initial_w should be None or a 2D numpy array.
		initial_b should be None or a 1D numpy array.
		'''
		super(InitializableDense, self).__init__()
		self.units = units
		self.with_bias = bias
		self.w_initializer = 'zeros' if initial_w is None else tf.constant_initializer(initial_w)

		if self.with_bias:
			self.b_initializer = 'zeros' if initial_b is None else tf.constant_initializer(initial_b)


	def build(self, input_shape):
		''' '''
		self.w = self.add_weight(shape=(input_shape[-1], self.units), \
			initializer=self.w_initializer, trainable=True, name='quad_w')

		if self.with_bias:
			self.b = self.add_weight(shape=(self.units,), \
				initializer=self.b_initializer, trainable=True, name='quad_b')


	def call(self, inputs):
		''' '''
		return tf.matmul(inputs, self.w)+self.b if self.with_bias else tf.matmul(inputs, self.w)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pandas as pd

from kxy.pre_learning import data_valuation as dv
from kxy.pre_learning import variable_selection as vs

from .base_accessor import BaseAccessor

@pd.api.extensions.register_dataframe_accessor("kxy_pre_learning")
class PreLearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for **post-learning** in supervised learning problems.

	This class defines the :code:`kxy_pre_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_pre_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""
	def data_valuation(self, target_column, problem_type=None, anonymize=None, snr='auto', include_mutual_information=False, file_name=None):
		"""
		Estimate the highest performance metrics achievable when predicting the :code:`target_column` using all other columns.

		When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`target_column` is categorical.


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		anonymize : None | bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost). When set to None (the default), your data will be anonymized when it is too big.
		include_mutual_information : bool
			Whether to include the mutual information between target and explanatory variables in the result.


		Returns
		-------
		achievable_performance : pandas.Dataframe
			The result is a pandas.Dataframe with columns (where applicable):

			* :code:`'Achievable Accuracy'`: The highest classification accuracy that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable R^2'`: The highest :math:`R^2` that can be achieved by a model using provided inputs to predict the label.
			* :code:`'Achievable RMSE'`: The lowest Root Mean Square Error that can be achieved by a model using provided inputs to predict the label.		
			* :code:`'Achievable Log-Likelihood Per Sample'`: The highest true log-likelihood per sample that can be achieved by a model using provided inputs to predict the label.



		.. admonition:: Theoretical Foundation

			Section :ref:`1 - Achievable Performance`.


		.. seealso::

			:ref:`kxy.pre_learning.achievable_performance.data_valuation <data-valuation>`
			
		"""
		assert target_column in self._obj.columns, 'The target_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'
		self.check_problem_type(problem_type, target_column)

		_obj = self.anonymize(columns_to_exclude=[target_column]) if anonymize or (anonymize is None and self.is_too_large) else self._obj

		return dv(_obj, target_column, problem_type, snr=snr, include_mutual_information=include_mutual_information, \
			file_name=file_name)


	def variable_selection(self, target_column, problem_type=None, anonymize=None, snr='auto', file_name=None):
		"""
		Runs the model-free variable selection analysis.

		When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`target_column` is categorical.


		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		anonymize : None | bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost). When set to None (the default), your data will be anonymized when it is too big.

		Returns
		-------
		result : pandas.DataFrame
			The result is a pandas.DataFrame with columns (where applicable):

			* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
			* :code:`'Variable'`: The column name corresponding to the input variable.
			* :code:`'Running Achievable R^2'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
			* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
			* :code:`'Running Achievable RMSE'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.


		.. admonition:: Theoretical Foundation

			Section :ref:`2 - Variable Selection Analysis`.

		.. seealso::

			:ref:`kxy.pre_learning.variable_selection.variable_selection <variable-selection>`
		"""
		assert target_column in self._obj.columns, 'The target_column should be a column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'
		self.check_problem_type(problem_type, target_column)
		
		_obj = self.anonymize(columns_to_exclude=[target_column]) if anonymize or (anonymize is None and self.is_too_large) else self._obj

		return vs(_obj, target_column, problem_type, snr=snr, file_name=file_name)








#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom Tensorflow losses.
"""
from multiprocessing import Pool, cpu_count
import numpy as np

import tensorflow as tf
tf.keras.backend.set_floatx('float64')
tf.config.threading.set_inter_op_parallelism_threads(2)
tf.config.threading.set_intra_op_parallelism_threads(8)
tf.config.set_soft_device_placement(True)
from tensorflow.python.ops import math_ops
from tensorflow.keras.losses import Loss


class MINDLoss(Loss):  
	'''
	Loss function.
	'''
	def call(self, y_true, y_pred):
		''' '''
		p_samples = y_pred[:, 0]
		q_samples = y_pred[:, 1]
		mi = -tf.reduce_mean(p_samples) + math_ops.log(tf.reduce_mean(math_ops.exp(q_samples)))
		return mi



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from time import time
import logging
import numpy as np

from tensorflow.keras.callbacks import EarlyStopping, TerminateOnNaN
from tensorflow.keras.optimizers import Adam

from kxy.misc.tf import PFSLearner, PFSOneShotLearner




def learn_principal_direction(y, x, ox=None, oy=None):
	"""
	Learn the i-th principal feature when using :math:`x` to predict :math:`y`.

	Parameters
	----------
	x : np.array
		2D array of shape :math:`(n, d)` containing original features.
	y : np.array
		Array of shape :math:`(n)` or :math:`(n, 1)` containing targets.

	Returns
	-------
	w : np.array
		The first principal direction.
	mi: float
		The mutual information :math:`I(y; w_i^Tx, \\dots, w_1^Tx)`.
	"""
	dx = 1 if len(x.shape) == 1 else x.shape[1]
	dy = 1 if len(y.shape) == 1 else y.shape[1]
	dox = 0 if ox is None else 1 if len(ox.shape) == 1 else ox.shape[1]
	doy = 0 if oy is None else 1 if len(oy.shape) == 1 else oy.shape[1]

	learner = PFSLearner(dx, dy=dy, dox=dox, doy=doy)
	learner.fit(x, y, ox=ox, oy=oy)

	mi = learner.mutual_information
	w = learner.feature_direction
	ox = learner.fx
	oy = learner.gy

	return w, mi, ox, oy



def learn_principal_directions_one_shot(y, x, p):
	"""
	Jointly learn p principal features.

	Parameters
	----------
	x : np.array
		2D array of shape :math:`(n, d)` containing original features.
	y : np.array
		Array of shape :math:`(n)` or :math:`(n, 1)` containing targets.
	p : int
		The number of principal features to learn.

	Returns
	-------
	w : np.array
		The matrix whose rows are the p principal directions.
	"""
	dx = 1 if len(x.shape) == 1 else x.shape[1]
	learner = PFSOneShotLearner(dx, p=p)
	learner.fit(x, y)
	w = learner.feature_directions

	return w




class PFS(object):
	def fit(self, x, y, p=None, mi_tolerance=0.0001, max_duration=None):
		"""
		Perform Principal Feature Selection using :math:`x` to predict :math:`y`.

		Specifically, we are looking for a :math:`p x d` matrix :math:`W` whose :math:`p` rows are learned sequentially such that :math:`z := Wx` is a great feature vector for predicting :math:`y`.

		Each row of :math:`W` is normal: :math:`||w_i||=1`, and the corresponding principal feature, namely :math:`w_i^Tx`, points in the same direction as :math:`y` (i.e. :math:`Cov(y, w_i^Tx) > 0`).

		The first row :math:`w_1` is learned so as to maximize the mutual information :math:`I(y; x^Tw_1)`.

		The second row :math:`w_2` is learned so as to maximize the conditional mutual information :math:`I(y; x^Tw_2 | x^Tw_1)`.	

		More generally, the :math:`(i+1)`-th row :math:`w_{i+1}` is learned so as to maximize the conditional mutual information :math:`I(y; x^Tw_{i+1} | [x^Tw_1, ..., x^Tw_i])`.


		Parameters
		----------
		x : np.array
			2D array of shape :math:`(n, d)` containing original features.
		y : np.array
			Array of shape :math:`(n)` or :math:`(n, 1)` containing targets.
		p : int | None (default)
			The number of features to select. When :code:`None` (the default) we stop when the estimated mutual information smaller than the mutual information tolerance parameter, or when we have exceeded the maximum duration. A value of :code:`p` that is not :code:`None` triggers one-shot PFS.
		mi_tolerance: float
			The smallest estimated mutual information required to keep looking for new feature directions.
		max_duration : float | None (default)
			The maximum amount of time (in second) to allocate to PFS.


		Returns
		-------
		W : np.array
			2D array whose rows are directions to use to compute principal features: :math:`z = Wx`.
		"""
		if max_duration:
			start_time = time()

		rows = []
		d = 1 if len(x.shape) == 1 else x.shape[1]
		if p is None:
			t = y.flatten().copy()
			old_mi = 0.0
			ox = None
			oy = None
			for i in range(d):
				w, mi, ox, oy = learn_principal_direction(t, x, ox=ox, oy=oy)

				if mi-old_mi < mi_tolerance:
					logging.info('The mutual information %.4f after %d round has not increase by more than %.4f: stopping.' % (
						mi, i+1, mi_tolerance))
					break
				else:
					logging.info('The mutual information has increased from %.4f to %.4f after %d rounds.' % (old_mi, mi, i+1))
					rows += [w.copy()]

				if max_duration:
					if time()-start_time > max_duration:
						logging.info('PFS has exceeded the configured maximum duration: exiting.')
						break

				old_mi = mi

			if rows == []:
				logging.warning('The only principal feature selected is not informative about the target: I(y; w^Tx)=%.4f' % mi)
				rows += [w.copy()]

			self.feature_directions = np.array(rows)
		else:
			# Learn all p principal features jointly.
			self.feature_directions = learn_principal_directions_one_shot(y, x, p)

		return self.feature_directions



class PCA(object):
	def __init__(self, energy_loss_frac=0.05):
		self.energy_loss_frac = energy_loss_frac


	def fit(self, x, _, max_duration=None, p=None):
		"""
		"""
		cov_x = np.cov(x.T) # Columns in x should represent variables and rows observations.
		u, d, v = np.linalg.svd(cov_x)
		cum_energy = np.cumsum(d)
		energy = cum_energy[-1]
		p = len([_ for _ in cum_energy if _ <= (1.-self.energy_loss_frac)*energy])

		self.feature_directions = u[:, :p].T

		return self.feature_directions







#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom Tensorflow models.
"""
import numpy as np
import tensorflow as tf
tf.keras.backend.set_floatx('float64')
tf.config.threading.set_inter_op_parallelism_threads(2)
tf.config.threading.set_intra_op_parallelism_threads(8)
tf.config.set_soft_device_placement(True)

from tensorflow.keras import Model
from tensorflow.keras.backend import clip
from tensorflow.keras.layers import Dense, Lambda, concatenate, Dot
from tensorflow.keras.constraints import UnitNorm

from .layers import InitializableDense


class CopulaModel(Model):
	"""
	Maximum-entropy copula under (possibly sparse) Spearman rank correlation constraints.
	"""
	def __init__(self, d, subsets=[]):
		super(CopulaModel, self).__init__()
		self.d = d
		if subsets == []:
			subsets = [[_ for _ in range(d)]]

		self.subsets = subsets
		self.n_subsets = len(self.subsets)
		self.p_samples = Lambda(lambda x: x[:,:,0])
		self.q_samples = Lambda(lambda x: x[:,:,1])

		self.fx_non_mon_layer_1s = [Dense(3, activation=tf.nn.relu) for _ in range(self.n_subsets)]
		self.fx_non_mon_layer_2s = [Dense(5, activation=tf.nn.relu) for _ in range(self.n_subsets)]
		self.fx_non_mon_layer_3s = [Dense(3, activation=tf.nn.relu) for _ in range(self.n_subsets)]
		self.fx_non_mon_layer_4s = [Dense(1) for _ in range(self.n_subsets)]

		eff_ds = [len(subset)+1 for subset in self.subsets]
		self.spears = [InitializableDense(eff_d) for eff_d in eff_ds]
		self.dots = [Dot(1) for _ in range(self.n_subsets)]

		# Mixing layers
		self.mixing_layer1 = Dense(5, activation=tf.nn.relu)
		self.mixing_layer2 = Dense(5, activation=tf.nn.relu)
		self.mixing_layer3 = Dense(1)


	def subset_statistics(self, u, i):
		'''
		Statistics function for the i-th subset of variables.
		'''
		n = tf.shape(u)[0]
		res = tf.zeros(shape=[n, 1], dtype=tf.float64)
		ui = tf.gather(u, self.subsets[i], axis=1)

		# Constraints beyond quadratic
		fui = self.fx_non_mon_layer_1s[i](ui)
		fui = self.fx_non_mon_layer_2s[i](fui)
		fui = self.fx_non_mon_layer_3s[i](fui)
		fui = self.fx_non_mon_layer_4s[i](fui)
		ui = concatenate([ui, fui], axis=1)
	
		# Spearman terms
		spearman_term = self.spears[i](ui)
		spearman_term = self.dots[i]([spearman_term, ui])
		res = tf.add(res, spearman_term)
		return res


	def statistics(self, u):
		'''
		Statistics function.
		''' 
		if self.n_subsets > 1:
			ts = [self.subset_statistics(u, i) for i in range(self.n_subsets)]
			t = concatenate(ts, axis=1)
			t = self.mixing_layer1(t)
			t = self.mixing_layer2(t)
			t = self.mixing_layer3(t)
		else:
			t = self.subset_statistics(u, 0)
		return t


	def call(self, inputs):
		'''
		'''
		p_samples = self.p_samples(inputs)
		t_p = self.statistics(p_samples)

		q_samples = self.q_samples(inputs)
		t_q = self.statistics(q_samples)
		
		t = concatenate([t_p, t_q], axis=1)
		t = clip(t, -100., 100.)
		return t


	def copula(self, inputs):
		'''
		'''
		u = tf.constant(inputs)
		c = math_ops.exp(self.statistics(u))
		return c.numpy()/c.numpy().mean()




class PFSOneShotModel(Model):
	"""
	Model for Principal Feature Selection.
	"""
	def __init__(self, x_ixs, y_ixs, ox_ixs=[], oy_ixs=[], p=1):
		super(PFSOneShotModel, self).__init__()
		self.p_samples = Lambda(lambda x: x[:,:,0])
		self.q_samples = Lambda(lambda x: x[:,:,1])

		# Feature direction
		self.w_constraint = UnitNorm(axis=0)
		self.w_layer = Dense(p, use_bias=False, kernel_constraint=self.w_constraint)

		# Z network
		n_outer_z = 3 # max(len(x_ixs)//10, 3)
		n_inner_z = 5 # max(len(x_ixs)//10, 5)
		self.z_layer_1 = Dense(n_outer_z, activation=tf.nn.relu)
		self.z_layer_2 = Dense(n_inner_z, activation=tf.nn.relu)
		self.z_layer_3 = Dense(n_outer_z, activation=tf.nn.relu)

		# Y network
		n_outer_y = 3 # max(len(x_ixs)//100, 3)
		n_inner_y = 5 # max(len(x_ixs)//100, 5)
		self.y_layer_1 = Dense(n_outer_y, activation=tf.nn.relu)
		self.y_layer_2 = Dense(n_inner_y, activation=tf.nn.relu)
		self.y_layer_3 = Dense(n_outer_y, activation=tf.nn.relu)

		# Outer quadratic constraints
		n_q = 1+n_outer_y+len(oy_ixs)+len(y_ixs)+n_outer_z+len(ox_ixs)
		self.quad_load = InitializableDense(n_q)
		self.quad_dot = Dot(1)

		# Indices of x and y.
		self.x_ixs = x_ixs
		self.y_ixs = y_ixs
		self.ox_ixs = ox_ixs
		self.oy_ixs = oy_ixs


	def fx(self, u):
		''' '''
		x = tf.gather(u, self.x_ixs, axis=1)

		# z = w^Tx
		z = self.w_layer(x)

		# f(z)
		fx = self.z_layer_1(z)
		fx = self.z_layer_2(fx)
		fx = self.z_layer_3(fx)

		if self.ox_ixs:
			old_fx = tf.gather(u, self.ox_ixs, axis=1)
			fx = concatenate([fx, old_fx], axis=1)

		return fx


	def gy(self, u):
		''' '''
		y = tf.gather(u, self.y_ixs, axis=1)
		gy = self.y_layer_1(y)
		gy = self.y_layer_2(gy)
		gy = self.y_layer_3(gy)

		if self.oy_ixs:
			old_gy = tf.gather(u, self.oy_ixs, axis=1)
			gy = concatenate([gy, old_gy], axis=1)

		return gy


	def statistics(self, u):
		'''
		'''
		n = tf.shape(u)[0]
		y = tf.gather(u, self.y_ixs, axis=1)
		res = tf.zeros(shape=[n, 1], dtype=tf.float64)

		fx = self.fx(u)
		gy = self.gy(u)

		# q = [1, f(z), y, g(y)]H[1, f(z), y, g(y)]^T
		# Note: the intercept is critical here, do not remove it!
		fxgy = concatenate([tf.ones(shape=[n, 1], dtype=tf.float64), fx, y, gy], axis=1)
		q = self.quad_load(fxgy)
		q = self.quad_dot([q, fxgy])

		res = tf.add(res, q)

		return res


	def call(self, inputs):
		'''
		'''
		p_samples = self.p_samples(inputs)
		t_p = self.statistics(p_samples)

		q_samples = self.q_samples(inputs)
		t_q = self.statistics(q_samples)
		
		t = concatenate([t_p, t_q], axis=1)
		t = clip(t, -200., 200.)
		return t



class PFSModel(PFSOneShotModel):
	"""
	Model for Principal Feature Selection.
	"""
	def __init__(self, x_ixs, y_ixs, ox_ixs=[], oy_ixs=[]):
		super(PFSModel, self).__init__(x_ixs, y_ixs, ox_ixs=ox_ixs, oy_ixs=oy_ixs, p=1)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import pickle as pkl

from .pfs_selector import PFS, PCA


class PFSPredictor(object):
	def _predict(self, obj):
		assert hasattr(self, 'models'), 'The model should be first fitted'
		assert hasattr(self, 'feature_directions'), 'The model should first be fitted'
		assert hasattr(self, 'x_columns'), 'The model should first be fitted'
		assert self.feature_directions.shape[0] > 0, 'There should be at least one feature selected'

		z = np.dot(obj[self.x_columns].values, self.feature_directions.T)
		y = self.models[0].predict(z)
		predictions = pd.DataFrame(index=obj.index)
		predictions[self.target_column] = y

		return predictions


	def predict(self, obj, memory_bound=False):
		"""
		Make predictions using the fitted model.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing test explanatory variables/features about which we want to make predictions.
		memory_bound : bool (Default False)
			Whether we should try to save memory.


		Returns
		-------
		result : pandas.DataFrame
			A dataframe with the same index as :code:`obj`, and with one column whose name is the :code:`target_column` used for training.
		"""
		if memory_bound:
			n = obj.shape[0]
			max_n = 1000000
			res = pd.DataFrame(index=obj.index)
			res[self.target_column] = np.nan
			i = 0
			while i < n:
				res.iloc[i:i+max_n] = self._predict(obj.iloc[i:i+max_n])
				i += max_n
			return res

		else:
			return self._predict(obj)


	def save(self, path):
		"""
		Cache the predictor to disk.
		"""
		meta_path = path + '-meta-' + self.__class__.__name__
		meta = {'target_column': self.target_column, 'feature_directions': self.feature_directions, 'x_columns': self.x_columns}
		with open(meta_path, 'wb') as f:
			pkl.dump(meta, f)
		self.models[0].save(path + '-' + self.__class__.__name__)
		

	@classmethod
	def load(cls, path, learner_func):
		"""
		Load the predictor from disk.
		"""
		meta_path = path + '-meta-' + cls.__name__
		with open(meta_path, 'rb') as f:
			meta = pkl.load(f)
		target_column = meta['target_column']
		feature_directions = meta['feature_directions']
		x_columns = meta['x_columns']

		n_vars = feature_directions.shape[0]
		model = learner_func(n_vars=n_vars, path=path + '-' + cls.__name__, safe=False)
		
		predictor = cls()
		predictor.models = [model]
		predictor.feature_directions = feature_directions
		predictor.target_column = target_column
		predictor.x_columns = x_columns

		return predictor


	def get_feature_selector(self):
		"""
		"""
		return PFS()

	@property
	def p(self):
		return self.feature_directions.shape[0]


	def fit(self, obj, target_column, learner_func, max_duration=None, path=None, p=None):
		"""
		Fits a supervised learner enriched with feature selection using the Principal Feature Selection (PFS) algorithm.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing training explanatory variables/features as well as the target.
		target_column : str
			The name of the column in :code:`obj` containing targets.
		learner_func : func | callable
			Function or callable that expects one optional argument :code:`n_vars` and returns an instance of a superviser learner (regressor or classifier) following the scikit-learn convention, and expecting :code:`n_vars` features. Specifically, the learner should have a :code:`fit(x_train, y_train)` method. The learner should also have a :code:`feature_importances_` property or attribute, which is an array or a list containing feature importances once the model has been trained. There should be as many importance scores in :code:`feature_importances_` as columns in :code:`fit(x_train, y_train)`.
		max_duration : float | None (default)
			If not None, then feature elimination will stop after this many seconds.
		p : int | None (default)
			The number of principal features to learn when using one-shot PFS.


		Attributes
		----------
		feature_directions : np.array
			The matrix whose rows are the directions in which to project the original features to get principal features.
		target_column : str
			The name of the column used as target.
		models : list
			An array whose first entry is the fitted model.
		x_columns : list
			The list of columns used for PFS sorted alphabetically.


		Returns
		-------
		results : dict
			A dictionary containing, among other things, feature directions.

		"""
		self.target_column = target_column
		self.x_columns = sorted([_ for _ in obj.columns if _ != target_column])

		x = obj[self.x_columns].values
		y = obj[[target_column]].values

		# Construct principal features
		principal_feature_selector = self.get_feature_selector()
		self.feature_directions = principal_feature_selector.fit(x, y, max_duration=max_duration, p=p)
		z = np.dot(x, self.feature_directions.T) # Principal features

		# Train the learner
		n_vars = self.feature_directions.shape[0]
		m = learner_func(n_vars=n_vars)
		m.fit(z, y)
		self.models = [m]
		if path:
			self.save(path)

		results = {'Feature Directions': self.feature_directions}
		return results


class PCAPredictor(PFSPredictor):
	def __init__(self, energy_loss_frac=0.05):
		self.energy_loss_frac = energy_loss_frac

	def get_feature_selector(self):
		return PCA(energy_loss_frac=self.energy_loss_frac)

